#include "ftpserver.h"
#include "config.h"
#include "netdiscovery.h"
#include "serverthread.h"
#include "stats.h"
#include <QTcpServer>
#include <QDebug>
#include "globals.h"

FtpServer *FtpServer::pInstance = NULL;

void FtpServer::startServer()
{
	if (pInstance)
		delete pInstance;
	pInstance = new FtpServer;
}

void FtpServer::stopServer()
{
	if (pInstance)
		delete pInstance;
	pInstance = NULL;
}

void FtpServer::restartServer()
{
	stopServer();
	startServer();
}

bool FtpServer::serverRunning()
{
	return pInstance != NULL;
}

FtpServer::FtpServer() : IPObject("qShareFtpServer", 42013), sock(NULL)
{
	pInstance = this;

	if (qshare::bGUImode)
	{
		connect(this, SIGNAL(newThread()), Stats::instance(), SLOT(updateStats()));
		connect(this, SIGNAL(closing()), Stats::instance(), SLOT(updateStats()));
		connect(this, SIGNAL(removeThread()), Stats::instance(), SLOT(updateStats()));
	}
	connect(this, SIGNAL(newThread()), this, SLOT(broadcastNewThread()));
	connect(this, SIGNAL(removeThread()), this, SLOT(broadcastRemoveThread()));
	connect(this, SIGNAL(enterServerMode()), this, SLOT(startService()));
	connect(this, SIGNAL(enterClientMode()), this, SLOT(startService()));
	connect(this, SIGNAL(clientLeft()), this, SLOT(removeClient()));

	totalSent = 0;
	totalReceived = 0;
	maxConnections = 0;
	totalConnections = 0;

	startService();
}

void FtpServer::removeClient()
{
	users.clear();
	validIDs.clear();
	send("refresh user list");
	users.insert(qshare::localUserName);
	validIDs << Config::getUID();
}

void FtpServer::startService()
{
	if (isServing() && sock == NULL)
	{
		sock = new QTcpServer(this);
		connect(sock,SIGNAL(newConnection()),this,SLOT(addConnection()));
		sock->listen(QHostAddress::Any, Config::getServerPort());
		users.clear();
		users.insert(qshare::localUserName);
		validIDs.clear();
		validIDs << Config::getUID();
	}
	else
		send("user", qshare::localUserName);
}

FtpServer::~FtpServer()
{
	emit closing();
	pInstance = NULL;
}

QList<FtpServer::ConnectionInfo> FtpServer::getInfo()
{
	QList<ConnectionInfo> lInfo;
	foreach(QObject *obj, children())
	{
		ServerThread *pThread = dynamic_cast<ServerThread*>(obj);
		if (!pThread)
			continue;
		ConnectionInfo info;
		info.user = pThread->getLogin();
		info.address = pThread->getAddress();
		switch(pThread->getMode())
		{
		case ServerThread::NONE:
			info.activity.clear();
			break;
		case ServerThread::RETR:
			info.activity = "RETR " + pThread->getFilename();
			break;
		case ServerThread::STOR:
			info.activity = "STOR " + pThread->getFilename();
			break;
		case ServerThread::APPE:
			info.activity = "APPE " + pThread->getFilename();
			break;
		case ServerThread::FIND:
			info.activity = "FIND " + pThread->getFilename();
			break;
		};
		lInfo << info;
	}
	return lInfo;
}

void FtpServer::addConnection()
{
	while (sock->hasPendingConnections())
	{
		QTcpSocket *s = sock->nextPendingConnection();
		if (s)
		{
			ServerThread *thread = new ServerThread(s, this);
			connect(thread,SIGNAL(destroyed()),this,SLOT(killThread()));
			++totalConnections;
		}
	}
	maxConnections = qMax(maxConnections, nbConnections());
	emit newThread();
}

void FtpServer::killThread()
{
	emit removeThread();
}

int FtpServer::nbConnections()
{
	int n(0);
	QObjectList list = children();
	foreach(QObject *obj, list)
	{
		if (dynamic_cast<ServerThread*>(obj))
			++n;
	}
	return n;
}

QList<FtpServer::ConnectionInfo> FtpServer::getClientsInfo()
{
	if (pInstance)
		return pInstance->getInfo();
	return QList<ConnectionInfo>();
}

void FtpServer::statsSend(int n)
{
	if (isServing())
		send("statsSend", n);
	totalSent += n;
}

void FtpServer::statsRecv(int n)
{
	if (isServing())
		send("statsRecv", n);
	totalReceived += n;
}

QStringList FtpServer::getServerInfo()
{
	QStringList lInfo;
	if (!pInstance)
	{
		lInfo << QString();
		lInfo << QString();
		lInfo << QString();
		lInfo << QString();
		lInfo << QString();
		return lInfo;
	}
	lInfo << QString::number(pInstance->totalSent);
	lInfo << QString::number(pInstance->totalReceived);
	lInfo << QString::number(pInstance->nbConnections());
	lInfo << QString::number(pInstance->maxConnections);
	lInfo << QString::number(pInstance->totalConnections);
	return lInfo;
}

void FtpServer::receive(const QByteArray &message)
{
	QDataStream stream(message);
	QString type;
	stream >> type;
	if (type == "SIG newThread")
	{
		if (!isServing())
			emit newThread();
		return;
	}
	if (type == "SIG removeThread")
	{
		if (!isServing())
			emit removeThread();
		return;
	}
	if (type == "statsSend")
	{
		int n;
		stream >> n;
		totalSent += n;
		return;
	}
	if (type == "statsRecv")
	{
		int n;
		stream >> n;
		totalReceived += n;
		return;
	}
	if (type == "user")
	{
		QString u;
		stream >> u;
		users << u;
		return;
	}
	if (type == "userid")
	{
		UserID uid;
		stream >> uid;
		validIDs << uid;
		return;
	}
	if (type == "refresh user list")
	{
		send("user", qshare::localUserName);
		send("userid", Config::getUID());
		return;
	}
}

void FtpServer::broadcastNewThread()
{
	if (isServing())
		send("SIG newThread");
}

void FtpServer::broadcastRemoveThread()
{
	if (isServing())
		send("SIG removeThread");
}

FtpServer *FtpServer::instance()
{
	return pInstance;
}

bool FtpServer::knownLogin(const QString &login) const
{
	for(QSet<UserID>::const_iterator it = validIDs.begin() ; it != validIDs.end() ; ++it)
	{
		if (it->getLogin() == login)
			return true;
	}
	return false;
}
